/*
 * Token replacement routine.
 *
 * Copyright 1999 by Gray Watson
 *
 * Permission to use, copy, modify, and distribute this software for
 * any purpose and without fee is hereby granted, provided that the
 * above copyright notice and this permission notice appear in all
 * copies, and that the name of Gray Watson not be used in advertising
 * or publicity pertaining to distribution of the document or software
 * without specific, written prior permission.
 *
 * Gray Watson makes no representations about the suitability of the
 * software described herein for any purpose.  It is provided "as is"
 * without express or implied warranty.
 *
 * The author may be contacted at http://256.com/gray/
 *
 * $Id: replace.h,v 1.3 2000/03/07 21:49:40 gray Exp $
 */

#ifndef __REPLACE_H__
#define __REPLACE_H__

#ifdef __cplusplus
extern "C" {
#endif

/*
 * Structure which describes a token -> value pair.  In the replace
 * format, a %token gets replaced in the output buffer with is
 * associated value.  This replacement array is terminated with an
 * entry where the token is a NULL pointer.
 */
typedef struct {
  char		*re_token;		/* token to replace [a-zA-Z0-9_]+ */
  char		*re_value;		/* value to replace the token with */
} replace_t;

/*<<<<<<<<<<  The below prototypes are auto-generated by fillproto */

/*
 * int replace_tokens
 *
 * DESCRIPTION:
 *
 * Take a input format with $token type tokens and a list of "token",
 * "value" pairs and insert the values where the tokens are in the
 * output buffer.
 *
 * For instance, if we have a format string: "ssh -l $user $host" and
 * a replacement array with { "user", "gray" }, { "host",
 * "gate.foo.com" } then the output buffer (if long enough) will
 * contain "ssh -l gray gate.foo.com".  The token name starts after $
 * and ends with a character not in the set "[a-zA-Z0-9_]".  You can
 * also use parentheses around the token if it does not end with a
 * space or tab.  For instance: "ssh -l $user $(host)00.$(domain)".
 * If "user" was "gray", "host" was server, and "domain" was "foo.com"
 * then the output buffer will contain "ssh -l gray server00.foo.com".
 * For this reason, parentheses cannot be contained in the token name
 * in the replacements array.  However, any characters including
 * parens can be in the replacement array values.
 *
 * NOTE: the output buffer will be \0 terminated only if there is
 * enough room for the \0 character.
 *
 * NOTE: if a token is in the format and not in the replacements
 * array, it will replaced with the empty string "".
 *
 * RETURNS:
 *
 * Number of characters added to the output buffer.
 *
 * ARGUMENTS:
 *
 * format -> Format with $tokens which describe what gets written into
 * the output buffer.  To specify a $ character which you want to make
 * it into the output buffer, you need to specify $$ like printf.
 *
 * format_len -> Length of the format string or -1 if it ends with a
 * \0.
 *
 * replacements -> Array of replace_t structures which list the tokens
 * and their associated values.  The array is terminated with an entry
 * where the token is a NULL pointer.
 *
 * output_buf <- Output buffer provided by the caller which will be
 * filled with characters from the format and the replacement values.
 *
 * output_buf_size -> Size of the output buffer.
 */
extern
int	replace_tokens(const char *format, const int format_len,
		       const replace_t *replacements, char *output_buf,
		       const int output_buf_size);

/*<<<<<<<<<<   This is end of the auto-generated output from fillproto. */

#ifdef __cplusplus
}
#endif

#endif /* ! __REPLACE_H__ */
