/*----------------------------------------------------------------------
 *
 *	Copyright  1996  Lycos, Inc. All Rights Reserved
 *
 *----------------------------------------------------------------------
 *
 *	THIS PROGRAM BELONGS TO LYCOS, INC. IT IS CONSIDERED
 *	A TRADE SECRET AND IS NOT TO BE DIVULGED OR USED BY
 *	PARTIES WHO HAVE NOT RECEIVED WRITTEN AUTHORIZATION FROM
 *	LYCOS, INC.
 *
 *----------------------------------------------------------------------
 * $Id: snprintf.h,v 1.4 2000/03/07 21:43:02 gray Exp $
 *----------------------------------------------------------------------*/

#ifndef __SNPRINTF_H__
#define __SNPRINTF_H__

#include <stdarg.h>			/* for va_list and ... */

#ifdef __cplusplus
extern "C" {
#endif

/*<<<<<<<<<<  The below prototypes are auto-generated by fillproto */

/*
 * int vsnprintf_tf
 *
 * DESCRIPTION:
 *
 * Like vsprintf however adds the maximum number of characters that can
 * be added to the output string.  This should be used in _all_ cases
 * instead of sprintf.
 *
 * RETURNS:
 *
 * The number of characters added to the output buffer.
 *
 * ARGUMENTS:
 *
 * out_buf - The output buffer where the format and arguments will be
 * copied.  This can be NULL if you want it to just calculate how much
 * space you need to expand a certain format.
 *
 * out_size - The total size of the output buffer, including space
 * reserved for the '\0' character.
 *
 * trunc_bp - Pointer to an integer which (if not NULL) will get set with
 * 1 if the buffer did not have enough room to hold the format and was
 * truncated or 0 if it did have the room.
 *
 * format - The printf-like format string.
 *
 * ap - An pointer to a varargs argument stack.
 */
extern
int	vsnprintf_tf(char *out_buf, int out_size, int *trunc_bp,
		     const char *format, va_list ap);

#ifndef HAS_SNPRINTF
/*
 * int vsnprintf
 *
 * DESCRIPTION:
 *
 * Like vsprintf however adds the maximum number of characters that can
 * be added to the output string.  This should be used in _all_ cases
 * instead of sprintf.
 *
 * RETURNS:
 *
 * The number of characters added to the output buffer.
 *
 * ARGUMENTS:
 *
 * out_buf - The output buffer where the format and arguments will be
 * copied.  This can be NULL if you want it to just calculate how much
 * space you need to expand a certain format.
 *
 * out_size - The total size of the output buffer, including space
 * reserved for the '\0' character.
 *
 * format - The printf-like format string.
 *
 * ap - An pointer to a varargs argument stack.
 */
extern
int	vsnprintf(char *out_buf, int out_size, const char *format, va_list ap);
#endif /* ifndef HAS_SNPRINTF */

#ifndef HAS_SNPRINTF
/*
 * int snprintf
 *
 * DESCRIPTION:
 *
 * Like sprintf however adds the maximum number of characters that can be
 * added to the output string.  This should be used in _all_ cases
 * instead of sprintf.
 *
 * RETURNS:
 *
 * The number of characters added to the output buffer.
 *
 * ARGUMENTS:
 *
 * out_buf - The output buffer where the format and arguments will be
 * copied.  This can be NULL if you want it to just calculate how much
 * space you need to expand a certain format.
 *
 * out_size - The total size of the output buffer, including space
 * reserved for the '\0' character.
 *
 * format - The printf-like format string.
 *
 * ... - Variable arguments that correspond to the format.
 */
extern
int	snprintf(char *out_buf, const unsigned int out_size,
		 char const *format, ...)
#ifdef __GNUC__
  __attribute__ ((format (printf, 3, 4)))
#endif
;
#endif /* ifndef HAS_SNPRINTF */

/*
 * int snprintf_tf
 *
 * DESCRIPTION:
 *
 * Like sprintf however adds the maximum number of characters that can be
 * added to the output string.  This should be used in _all_ cases
 * instead of sprintf.
 *
 * RETURNS:
 *
 * The number of characters added to the output buffer.
 *
 * ARGUMENTS:
 *
 * out_buf - The output buffer where the format and arguments will be
 * copied.  This can be NULL if you want it to just calculate how much
 * space you need to expand a certain format.
 *
 * out_size - The total size of the output buffer, including space
 * reserved for the '\0' character.
 *
 * trunc_bp - Pointer to an integer which (if not NULL) will get set with
 * 1 if the buffer did not have enough room to hold the format and was
 * truncated or 0 if it did have the room.
 *
 * format - The printf-like format string.
 *
 * ... - Variable arguments that correspond to the format.
 */
extern
int	snprintf_tf(char *out_buf, const unsigned int out_size, int *trunc_bp,
		    char const *format, ...)
#ifdef __GNUC__
  __attribute__ ((format (printf, 4, 5)))
#endif
;

/*
 * lycos_snprintf
 *
 * DESCRIPTION:
 *
 * Alias for snprintf() to force the use of the snprintf() in this
 * library rather than any system-defined functions.  For example:
 * the Netscape Enterprise Server defines an snprintf() in its API
 * which is _broken_ on Solaris, so WPS calls this function instead.
 *
 * RETURNS:
 *
 * The number of characters added to the output buffer.
 *
 * ARGUMENTS:
 *
 * out_buf - The output buffer where the format and arguments will be
 * copied.  This can be NULL if you want it to just calculate how much
 * space you need to expand a certain format.
 *
 * out_size - The total size of the output buffer, including space
 * reserved for the '\0' character.
 *
 * format - The printf-like format string.
 *
 * ... - Variable arguments that correspond to the format.
 */
extern
int	lycos_snprintf(char *out_buf, const unsigned int out_size,
		 char const *format, ...)
#ifdef __GNUC__
  __attribute__ ((format (printf, 3, 4)))
#endif
;

/*<<<<<<<<<<   This is end of the auto-generated output from fillproto. */

#ifdef __cplusplus
}
#endif

#endif /* __SNPRINTF_H__ */
